@@ -1,5 +1,4 @@
-"""upathlib"""
-__version__ = '0.6.3'
+__version__ = '0.6.5'
 
 from ._upath import Upath, FileInfo, LockAcquisitionTimeoutError
 from ._local import LocalUpath
@@ -7,9 +6,7 @@ from ._blob import BlobUpath
 
 
 __all__ = [
-    'Upath',
+    'Upath', 'LocalUpath', 'BlobUpath',
     'FileInfo',
     'LockAcquisitionTimeoutError',
-    'LocalUpath',
-    'BlobUpath',
 ]
