@@ -12,18 +12,25 @@ import logging
 import os
 import os.path
 import pathlib
-import time
+import queue
+import threading
 from dataclasses import dataclass
 from io import UnsupportedOperation
-from typing import List, Iterator, Tuple, Type, TypeVar, Any, Optional, Union, Callable, cast
+from typing import List, Iterable, Iterator, Type, TypeVar, Any, Optional, Tuple, Callable
 
 from overrides import EnforceOverrides
+from tqdm import tqdm
 from .serializer import (
     ByteSerializer, TextSerializer,
-    JsonSerializer, PickleSerializer, CompressedPickleSerializer,
-    OrjsonSerializer, CompressedOrjsonSerializer,
+    JsonSerializer, ZJsonSerializer, ZstdJsonSerializer,
+    PickleSerializer, ZPickleSerializer, ZstdPickleSerializer,
+    OrjsonSerializer, ZOrjsonSerializer, ZstdOrjsonSerializer,
 )
 
+# User may want to do this:
+#  logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)
+# to suppress the "urllib3 connection lost" warning.
+# `upathlib` as a library does not make such changes.
 
 logger = logging.getLogger(__name__)
 T = TypeVar('T', bound='Upath')
@@ -43,145 +50,111 @@ class FileInfo:
     details: Any   # platform-dependent
 
 
-def _execute_in_thread_pool(
-        jobs,
-        concurrency: int = None,
-        *,
-        retry_on_exceptions: Optional[Union[Type[Exception],
-                                            Tuple[Type[Exception]]]] = None,
-        retries: int = 3,
-):
-    '''
-    If you want to skip certain errors and don't retry on them,
-    use `retries=0`.
-    '''
-    if concurrency is None:
-        concurrency = 16
-    else:
-        assert 0 <= concurrency <= 64
-
-    if concurrency == 0:
-        # TODO: retry on error.
-        results = []
-        for f, args, kwargs in jobs:
-            z = f(*args, **kwargs)
-            results.append(z)
-        return results
-
-    class MyExc(Exception):
-        pass
-
-    if retry_on_exceptions is None:
-        def ff(f, args, kwargs):
-            return f(*args, **kwargs)
-    else:
-        def ff(f, args, kwargs):
-            try:
-                return f(*args, **kwargs)
-            except retry_on_exceptions:
-                return MyExc(f, args, kwargs)
-
-    with concurrent.futures.ThreadPoolExecutor(concurrency) as pool:
-        results = []
-        tasks = []
-        bad = []
-        for args in jobs:
-            tasks.append(pool.submit(ff, *args))
-        for f in concurrent.futures.as_completed(tasks):
-            z = f.result()
-            if isinstance(z, MyExc):
-                bad.append(z.args)
-            else:
-                results.append(z)
-        jobs = bad
-
-    if not jobs:
-        return results
-
-    # Switch to sequential retry.
-    # Some issues could be due to large files.
-    for _ in range(retries):
-        logger.warning('failed on %d items; retrying', len(jobs))
-        time.sleep(0.6)
-        bad = []
-        for args in jobs:
-            z = ff(*args)
-            if isinstance(z, MyExc):
-                bad.append(z.args)
-            else:
-                results.append(z)
-        if not bad:
-            break
-        jobs = bad
-
-    if not jobs:
-        return results
-
-    logger.error(f'failed on {len(jobs)} items: {jobs}')
-    raise RuntimeError(f'failed on {len(jobs)} items')
-
-
-def _should_update(source: Upath, target: Upath) -> bool:
-    sourceinfo = cast(FileInfo, source.file_info())
-    targetinfo = cast(FileInfo, target.file_info())
-    return (sourceinfo.size != targetinfo.size
-            or sourceinfo.mtime > targetinfo.mtime)
-    # Otherwise, we're assuming that
-    # the target file was copied from the source
-    # previously.
-
-
 class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-methods
-    @staticmethod
-    def _should_overwrite(source: Upath,
-                          target: Upath,
-                          *,
-                          exist_action: str = None,
-                          update_filter: Callable[[Upath, Upath], bool] = None,
-                          ) -> bool:
-        # Determine whether `source` should overwrite `target`,
-        # both of which being existing files, usually of the same name
-        # in the context of 'copying' or 'exporting'.
-        if exist_action is None:
-            exist_action = 'raise'
-        else:
-            assert exist_action in ('raise', 'skip', 'overwrite', 'update')
+    _thread_executor_ = {
+        'nest0': concurrent.futures.ThreadPoolExecutor(
+            min(32, (os.cpu_count() or 1) + 4),
+            thread_name_prefix='UpathExecutor0'),
+        'nest1': concurrent.futures.ThreadPoolExecutor(
+            10,
+            thread_name_prefix='UpathExecutor1'),
+    }
+    _tqdm_bar_format_ = "{desc} | {percentage:3.0f}% | {n:.0f}/{total:.0f}, {elapsed}"
 
-        if exist_action == 'raise':
-            raise FileExistsError(target)
-        if exist_action == 'skip':
-            logger.info(f"target file {target!r} exists; skipped")
-            return False
-        if exist_action == 'update':
-            if update_filter is None:
-                update_filter = _should_update
-            if not update_filter(source, target):
-                logger.info(
-                    f"target {target!r} appears to be up-to-date; skipped")
-                return False
-        return True
+    @classmethod
+    def _run_in_executor(cls, tasks: Iterable[Tuple[Callable, tuple, dict, str]], description: str):
+        '''
+        This method is used to run multiple I/O jobs concurrently, e.g.
+        uploading/downloading all files in a folder recursively.
+        Where supported, this may also be used to download a large blob
+        by splitting the work into multiple segments.
+
+        `tasks`: each element is a tuple of (func, args, kwargs, description).
+        `description`: used at the beginning of the tqdm progress bar.
+        '''
+        if not isinstance(tasks, list):
+            tasks = list(tasks)
+        n_tasks = len(tasks)
+        if not n_tasks:
+            return
+
+        def enqueue(tasks, executor, q):
+            for func, args, kwargs, desc in tasks:
+                t = executor.submit(func, *args, **kwargs)
+                q.put((t, desc))
+            q.put(None)
+
+        if threading.current_thread().name.startswith('UpathExecutor0'):
+            executor = cls._thread_executor_['nest1']
+        else:
+            executor = cls._thread_executor_['nest0']
+
+        q = queue.Queue(executor._max_workers * 2)
+        task = threading.Thread(target=enqueue, args=(tasks, executor, q))
+        task.start()
+        with tqdm(total=n_tasks, bar_format=cls._tqdm_bar_format_) as pbar:
+            pbar.set_description(description)
+            while True:
+                z = q.get()
+                if z is None:
+                    break
+                t, desc = z
+                pbar.set_description(description + desc + ' ...')
+                pbar.update(0.5)
+                try:
+                    yield t.result()
+                except Exception:
+                    while True:
+                        z = q.get()
+                        if z is None:
+                            break
+                        t, desc = z
+                        t.cancel()
+                        # This may not succeed, but there isn't a good way to
+                        # guarantee cancellation here.
+                    raise
+                pbar.update(0.5)
+            pbar.set_description(description)
+            task.join()
 
     @classmethod
     def register_read_write_byte_format(cls, serde: Type[ByteSerializer], name: str):
-        def _write(self, data, *, overwrite: bool = False):
-            return self.write_bytes(serde.serialize(data), overwrite=overwrite)
+        '''
+        For example, if `serde` is a ByteSerializer subclass and `name` is 'myway',
+        then this method adds isinstance methods `write_myway` and `read_myway`.
 
-        def _read(self):
+        `name`: needs to be valid method name, e.g. can't contain space or dash.
+        '''
+        def _write(self, data, *, overwrite=False, **kwargs):
+            return self.write_bytes(
+                serde.serialize(data, **kwargs),
+                overwrite=overwrite)
+
+        def _read(self, **kwargs):
             z = self.read_bytes()
-            return serde.deserialize(z)
+            return serde.deserialize(z, **kwargs)
 
+        setattr(_write, '__name__', f'write_{name}')
+        setattr(_read, '__name__', f'read_{name}')
         setattr(cls, f'write_{name}', _write)
         setattr(cls, f'read_{name}', _read)
 
     @classmethod
     def register_read_write_text_format(cls, serde: Type[TextSerializer], name: str):
-        def _write(self, data, *, overwrite: bool = False):
-            return self.write_text(serde.serialize(data), overwrite=overwrite)
+        '''
+        Anologous to `register_read_write_byte_format`.
+        '''
+        def _write(self, data, *, overwrite=False, **kwargs):
+            return self.write_text(
+                serde.serialize(data, **kwargs),
+                overwrite=overwrite)
 
-        def _read(self):
+        def _read(self, **kwargs):
             z = self.read_text()
-            return serde.deserialize(z)
+            return serde.deserialize(z, **kwargs)
 
+        setattr(_write, '__name__', f'write_{name}')
+        setattr(_read, '__name__', f'read_{name}')
         setattr(cls, f'write_{name}', _write)
         setattr(cls, f'read_{name}', _read)
 
@@ -194,6 +167,11 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         no difference. The path constructed with `*pathsegments`
         is always "absolute" under a known "root".
 
+            Note that if one segment starts with '/', it will reset to the "root"
+            and discard all the segments that have come before it.
+
+            If missing, the path constructed is the "root".
+
         For a local POSIX file system, the root is the regular `/`.
         For Azure blob store, the root is that of a "container".
         For AWS and GCP blob stores, the root is that of a "bucket".
@@ -202,10 +180,12 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         self._path = os.path.normpath(os.path.join(
             '/', *pathsegments))  # pylint: disable=no-value-for-parameter
         # The path is always "absolute" starting with '/'.
-        # Unless it is `/`, it does not have a trailing `/`.
+        # It does not have a trailing `/` unless the path is just `/` itself.
 
     def __repr__(self) -> str:
         return f"{self.__class__.__name__}('{self._path}')"
+        # Subclass may want to customize this method to add more info,
+        # e.g. "bucket" name.
 
     def __str__(self) -> str:
         return self._path
@@ -214,6 +194,8 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         if other.__class__ is not self.__class__:
             return NotImplemented
         return self._path == other._path
+        # Subclass may want to customize this method to check more things,
+        # e.g. whether `self` and `other` are in the same "bucket".
 
     def __lt__(self, other) -> bool:
         if other.__class__ is not self.__class__:
@@ -239,16 +221,12 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         return hash(repr(self))
 
     def __truediv__(self: T, key: str) -> T:
+        '''
+        This is called by `self / key`.
+        '''
         return self.joinpath(key)
 
-    def copy_dir(self: T,
-                 target: str,
-                 *,
-                 concurrency: int = None,
-                 exist_action: str = None,
-                 update_filter: Callable[[Upath, Upath], bool] = None,
-                 **kwargs,
-                 ) -> int:
+    def copy_dir(self, target: str, *, overwrite: bool = False) -> int:
         '''Analogous to `copy_file`.
 
         Return the number of files copied.
@@ -262,25 +240,23 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                 extra = str(p.path.relative_to(self.path))
                 yield (
                     p.copy_file,
-                    [(target_ / extra)._path],
-                    {'exist_action': exist_action, 'update_filter': update_filter},
+                    ((target_ / extra)._path, ),
+                    {'overwrite': overwrite},
+                    extra,
                 )
 
-        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
-        return sum(nn)
+        n = 0
+        for _ in self._run_in_executor(foo(), f'copying dir {self.name}/'):
+            n += 1
+        return n
 
-    def _copy_file(self: T, target: T) -> None:
-        # `target` is a path in the same store, but does not exist.
+    def _copy_file(self: T, target: T, *, overwrite: bool = False) -> None:
+        # `target` is a path in the same store.
         # Reference implementation.
         # Subclass may customize this to perform file operations.
-        target.write_bytes(self.read_bytes())
+        target.write_bytes(self.read_bytes(), overwrite=overwrite)
 
-    def copy_file(self: T,
-                  target: str,
-                  *,
-                  exist_action: str = None,
-                  update_filter: Callable[[Upath, Upath], bool] = None,
-                  ) -> int:
+    def copy_file(self, target: str, *, overwrite: bool = False) -> None:
         '''Copy file to `target` in the same store.
 
         `target` is either absolute, or relative to `self.parent`.
@@ -289,43 +265,21 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
 
         If `self` is not an existing file, raise `FileNotFoundError`.
 
-        If `target` is an existing file, then the behavior depends on
-        `exist_action`:
-
-            'raise' (default): raise `FileExistsError`.
-            'skip': skip this file; proceed to work on other files.
-            'overwrite': overwrite the existing file.
-            'update': overwrite if source `mtime` is newer than target,
-                or source and target have diff size; otherwise skip.
+        If `target` is an existing file and `overwrite` is `False`,
+        raise `FileExistsError`.
 
         If `target` is an existing directory, raise `IsADirectoryError`.
         Note: this behavior is different from the Unix command `cp`
         in this situation---it does not *copy into* the target directory.
-
-        Return number of files copied (0 or 1).
         '''
-        # Reference implementation.
-        # Subclass should implement by direct file operation if possible.
-        if not self.is_file():
-            raise FileNotFoundError(self)
         target_ = self.parent / target
         if target_ == self:
-            return 0
-
-        if target_.is_file():
-            if self._should_overwrite(self, target_,
-                                      exist_action=exist_action,
-                                      update_filter=update_filter):
-                target_.remove_file()
-                self._copy_file(target_)
-                return 1
-            return 0
+            return
 
         if target_.is_dir():
             raise IsADirectoryError(target_)
 
-        self._copy_file(target_)
-        return 1
+        self._copy_file(target_, overwrite=overwrite)
 
     def exists(self) -> bool:
         '''Return `True` if the path is an existing file or dir,
@@ -343,26 +297,22 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         '''
         return self.is_file() or self.is_dir()
 
-    def export_dir(self,
-                   target: Upath,
-                   *,
-                   concurrency: int = None,
-                   exist_action: str = None,
-                   update_filter: Callable[[Upath, Upath], bool] = None,
-                   **kwargs,
-                   ) -> int:
-        '''Copy the content of `self` to the specified `target`,
-        which is typically in another store.
+    def export_dir(self, target: Upath, *, overwrite: bool = False) -> int:
+        '''Copy the content of the current directory recursively
+        to the specified `target`, which is typically in another store.
+
+        `target` corresponds to `self`, that is, direct children of `self`
+        are copied directly into `target`.
 
         Compare with `copy_dir`, which make copies within the same store.
 
-        `concurrency`: number of threads to use. If `None`,
-        a default value (e.g. 4) is used.
-
         Overwriting happens file-wise. For example, if the target directory
         contains files that do not exist in the source directory, they
         are left untouched.
 
+        If `target` is a `LocalUpath` object, then a subclass may implement more
+        efficient ways to "download", and also renaming this method to "download_dir".
+
         Return the number of files copied.
         '''
         def foo():
@@ -371,59 +321,41 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                 extra = str(p.path.relative_to(self_path))
                 yield (
                     p.export_file,
-                    [target / extra],
-                    {'exist_action': exist_action,
-                     'check_source_exist': False,
-                     'update_filter': update_filter},
+                    (target / extra, ),
+                    {'overwrite': overwrite},
+                    extra,
                 )
 
-        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
-        return sum(nn)
+        n = 0
+        for _ in self._run_in_executor(foo(), f'exporting dir {self.name}/'):
+            n += 1
+        return n
 
-    def _export_file(self, target: Upath) -> None:
-        # `target` is a non-existent path in another store.
+    def _export_file(self, target: Upath, *, overwrite: bool = False) -> None:
         # Reference implementation.
         # Subclass may customize this to perform file download
         # when `target` is a `LocalUpath`.
-        target.write_bytes(self.read_bytes())
+        target.write_bytes(self.read_bytes(), overwrite=overwrite)
 
-    def export_file(self,
-                    target: Upath,
-                    *,
-                    exist_action: str = None,
-                    update_filter: Callable[[Upath, Upath], bool] = None,
-                    check_source_exist: bool = True,
-                    ) -> int:
+    def export_file(self, target: Upath, *, overwrite: bool = False) -> None:
         '''Copy the file to the specified `target`, which is typically
         in another store.
 
-        Return the number of files copied (0 or 1).
-
-        The `target` specifies the name corresponding the the name of `self`.
-        If `target` is an existing directory, a `FileExistsError` is raised.
+        The `target` specifies a blob corresponding to `self`.
+        If `target` is an existing directory, a `IsADirectoryError` is raised.
         A copy is not placed *into* the target directory. This behavior
         differs from the Linux command `cp`.
 
-        Compare with `copy_file`, which make copies within the same store.
+        If `target` is a `LocalUpath` object, then a subclass may implement more
+        efficient ways to "download", and also renaming this method to "download_file".
+
+        Compare with `copy_file`, which makes copies within the same store.
         '''
-        if check_source_exist and not self.is_file():
-            raise FileNotFoundError(self)
-
-        if target.is_file():
-            if self._should_overwrite(self, target,
-                                      exist_action=exist_action,
-                                      update_filter=update_filter):
-                target.remove_file()
-            else:
-                return 0
-
         if target.is_dir():
             # Do not delete.
-            raise FileExistsError(target)
+            raise IsADirectoryError(target)
 
-        logger.info("copying '%s' to '%s'", self, target)
-        self._export_file(target)
-        return 1
+        self._export_file(target, overwrite=overwrite)
 
     @ abc.abstractmethod
     def file_info(self) -> Optional[FileInfo]:
@@ -432,14 +364,7 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         '''
         raise NotImplementedError
 
-    def import_dir(self,
-                   source: Upath,
-                   *,
-                   concurrency: int = None,
-                   exist_action: str = None,
-                   update_filter: Callable[[Upath, Upath], bool] = None,
-                   **kwargs,
-                   ) -> int:
+    def import_dir(self, source: Upath, *, overwrite: bool = False) -> int:
         '''Analogous to `export_dir`.
         '''
         def foo():
@@ -448,47 +373,26 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                 extra = str(p.path.relative_to(source_path))
                 yield (
                     (self / extra).import_file,
-                    [p],
-                    {'exist_action': exist_action,
-                     'check_source_exist': False,
-                     'update_filter': update_filter},
+                    (p, ),
+                    {'overwrite': overwrite},
+                    extra,
                 )
 
-        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
-        return sum(nn)
+        n = 0
+        for _ in self._run_in_executor(foo(), f'importing dir {self.name}/'):
+            n += 1
+        return n
 
-    def _import_file(self, source: Upath) -> None:
-        # `self` does not exist, hence no concern about overwriting.
-        # Subclass may customize this to perform file upload
-        # when `target` is a `LocalUpath`.
-        # This is not used by `import_file` directly, but
-        # it is used by `export_file` in certain situations.
-        # See `LocalUpath._export_file`.
-        self.write_bytes(source.read_bytes())
-
-    def import_file(self, source: Upath, *,
-                    exist_action: str = None,
-                    update_filter: Callable[[Upath, Upath], bool] = None,
-                    check_source_exist: bool = True,
-                    ) -> int:
-        if check_source_exist and not source.is_file():
-            raise FileNotFoundError(source)
-
-        if self.is_file():
-            if self._should_overwrite(source, self,
-                                      exist_action=exist_action,
-                                      update_filter=update_filter):
-                self.remove_file()
-            else:
-                return 0
+    def _import_file(self, source: Upath, *, overwrite: bool = False) -> None:
+        self.write_bytes(source.read_bytes(), overwrite=overwrite)
 
+    def import_file(self, source: Upath, *, overwrite: bool = False) -> None:
+        # When `target` is a `LocalUpath`, subclass may implement this
+        # in more efficient ways for uploading, and rename it to `upload_file`.
         if self.is_dir():
             # Do not delete.
-            raise FileExistsError(self)
-
-        logger.info("copying '%s' to '%s'", source, self)
-        self._import_file(source)
-        return 1
+            raise IsADirectoryError(self)
+        self._import_file(source, overwrite=overwrite)
 
     @ abc.abstractmethod
     def is_dir(self) -> bool:
@@ -532,7 +436,8 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         we say `/a/b/c` is a "file", and also a "dir".
         User is recommended to avoid such namings.
 
-        This situation does not happen in a local file system.'''
+        This situation does not happen in a local file system.
+        '''
         raise NotImplementedError
 
     @ abc.abstractmethod
@@ -543,9 +448,11 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         Each yielded element is either a file or a dir.
 
         If `self` is not a dir (e.g. maybe it's a file),
-        or does not exist at all, yield nothing, but do not raise exception.
+        or does not exist at all, yield nothing (resulting in an
+        empty iterable), but do not raise exception.
 
-        There is no guarantee on the order of the returned elements.'''
+        There is no guarantee on the order of the returned elements.
+        '''
         raise NotImplementedError
 
     def joinpath(self: T, *other: str) -> T:
@@ -558,16 +465,17 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         '''Lock the file pointed to, in order to have exclusive access.
 
         `timeout`: if the lock can't be acquired within *timeout* seconds,
-        raise `LockAcquisitionTimeoutError`. Default is waiting for ever.
-        Once a lease is acquired, it will not expire until this contexmanager
-        exits. In other word, this is timeout for the "wait", not for the
+        raise `LockAcquisitionTimeoutError`. If `None`, wait for ever until
+        a lock is acquired. Once a lease is acquired,
+        it will not expire until this contexmanager exits.
+        In other words, this is timeout for the "wait", not for the
         lease itself. Actual waiting time may be slightly longer.
 
         This is a "mandatory lock", as opposed to an "advisory lock".
         However, this API does not specify that the locked file
         can be accessed for its content or used in any particular way.
         The intended use case is for this lock to be used
-        in implementing a (cooperative) "code lock".
+        for implementing a (cooperative) "code lock".
 
         The `yield` statement is not required to yield anything,
         that is, it may be simply
@@ -583,7 +491,13 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
 
             f = Upath('abc.txt')
             with f.with_suffix('.txt.lock').lock():
-                ...  # now read/write `f` with exclusive access
+                ...
+                # now write `f` with exclusive access,
+                # because any other (cooperative) code block
+                # will not be able to get hold of `abc.txt.lock`
+                # in order to write `f` in its context-managed block.
+                # It's up to the program design whether this lock
+                # covers reading as well.
 
         Some storage engines may not provide the capability to implement
         this lock.
@@ -595,8 +509,10 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
 
     @ property
     def name(self) -> str:
-        '''Return the segment after the last `/`.'''
-        # If `self.path` is '/', then `self.path.name` is ''.
+        '''Return the segment after the last `/`.
+
+        If `self.path` is '/', then `self.path.name` is ''.
+        '''
         return self.path.name
 
     @ property
@@ -616,83 +532,81 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         '''
         raise NotImplementedError
 
-    def read_text(self, *, encoding: str = 'utf-8', errors: str = 'strict'):
+    def read_text(self):
         # Refer to https://docs.python.org/3/library/functions.html#open
-        return self.read_bytes().decode(encoding=encoding, errors=errors)
+        return self.read_bytes().decode(encoding='utf-8', errors='strict')
 
-    def remove_dir(self, *, concurrency: int = None, **kwargs) -> int:
+    # TODO: rename 'remove' to 'delete'?
+
+    def remove_dir(self) -> int:
         '''Remove the directory pointed to by `self`,
         along with all its contents, recursively.
 
         Return the number of files removed.
 
-        `concurrency`: number of threads to use. If `None`,
-        a default value is used.
-
         Local upath needs to customize this implementation, because
         it needs to take care of deleting "empty" subdirectories.
         '''
         def foo():
             for p in self.riterdir():
-                yield p.remove_file, [], {}
+                yield p.remove_file, [], {}, str(p.path.relative_to(self.path))
 
-        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
-        return sum(nn)
+        n = 0
+        for _ in self._run_in_executor(foo(), f'removing dir {self.name}/'):
+            n += 1
+        return n
 
     @abc.abstractmethod
-    def remove_file(self) -> int:
+    def remove_file(self) -> None:
         '''Remove the file pointed to by `self`.
 
-        Return the number of files removed (0 or 1).
-
-        If `self` is not an existing file, return 0.
-        If the file exists but can't be removed, raise an exception.
+        If `self` is not an existing file, raise FileNotFoundError.
+        If the file exists but can't be removed, usually the platform-dependent
+        exception is propagated.
         '''
         raise NotImplementedError
 
-    def rename_dir(self: T,
-                   target: str,
-                   *,
-                   concurrency: int = None,
-                   exist_ok: bool = False,
-                   **kwargs,
-                   ) -> T:
+    def rename_dir(self: T, target: str, *, overwrite: bool = False) -> T:
         '''Analogous to `rename_file`.
 
+        `overwrite` is applied per file, which suggests that if there are
+        files under `target` that do not have counterparts under `self`,
+        they are left untouched.
+
         Local upath needs to customize this implementation, because
         it needs to take care to delete empty subdirectories under `self`.
         '''
         if not self.is_dir():
-            raise FileNotFoundError(self)
+            raise FileNotFoundError(str(self))
 
         target_ = self.parent / target
         if target_ == self:
             return self
 
-        if not exist_ok:
-            if target_.exists():
-                raise FileExistsError(target_)
-
         def foo():
             for p in self.riterdir():
                 extra = str(p.path.relative_to(self.path))
                 yield (
                     p.rename_file,
                     [(target_ / extra)._path],
-                    {},
+                    {'overwrite': overwrite},
+                    extra,
                 )
 
-        _ = _execute_in_thread_pool(foo(), concurrency, **kwargs)
+        for _ in self._run_in_executor(foo(), f'renaming dir {self.name}/'):
+            pass
         return target_
 
-    def _rename_file(self: T, target: T):
-        '''Rename `self` to `target`, which is a non-existent path
-        in the same store.
+    def _rename_file(self: T, target: str, *, overwrite: bool = False):
+        '''Rename `self` to `target`, which is a path in the same store.
+
+        This is a reference implementation. There are likely
+        more efficient ways to do this on any specific platform.
         '''
-        self._copy_file(target)
+        self.copy_file(target, overwrite=overwrite)
         self.remove_file()
 
-    def rename_file(self: T, target: str) -> T:
+    def rename_file(self: T, target: str, *, overwrite: bool = False) -> T:
         '''Rename the current file to `target` in the same store.
 
         `target` is either absolute or relative to `self.parent`.
@@ -701,16 +615,11 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
 
         Return an object pointing to the new path.
         '''
-        if not self.is_file():
-            raise FileNotFoundError(self)
         target_ = self.parent / target
         if target_ == self:
             return self
 
-        if target_.exists():
-            raise FileExistsError(target_)
-
-        self._rename_file(target_)
+        self._rename_file(target_._path, overwrite=overwrite)
         return target_
 
     @abc.abstractmethod
@@ -724,19 +633,16 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         Similar to `iterdir`, if `self` is not a dir or does not exist,
         then nothing is yielded, and no exception is raised either.
 
-        There is no guarantee on the order of the returned elements.'''
-
+        There is no guarantee on the order of the returned elements.
+        '''
         raise NotImplementedError
 
-    def rmrf(self, *, concurrency: int = None, **kwargs) -> int:
+    def rmrf(self) -> int:
         '''Analogous to `rm -rf`. Remove the file or dir pointed to
         by `self`.
 
         Return the number of files removed.
 
-        `concurrency`: number of threads to use. If `None`,
-        a default value (e.g. 4) is used.
-
         For example, if these blobs are present:
 
             /a/b/c/d/e.txt
@@ -747,9 +653,17 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         '''
         if self._path == '/':
             raise UnsupportedOperation("`rmrf` not allowed on root directory")
-        n1 = self.remove_file()
-        n2 = self.remove_dir(concurrency=concurrency, **kwargs)
-        return n1 + n2
+        try:
+            self.remove_file()
+        except (FileNotFoundError, IsADirectoryError):
+            n = 0
+        else:
+            n = 1
+        try:
+            m = self.remove_dir()
+        except FileNotFoundError:
+            m = 0
+        return n + m
 
     @ property
     def stem(self) -> str:
@@ -793,11 +707,9 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
     def write_bytes(self,
                     data: bytes,
                     *,
-                    overwrite: bool = False) -> int:
+                    overwrite: bool = False) -> None:
         '''Write bytes to file.
 
-        Return number of bytes written.
-
         Parent directories are created as needed.
 
         `overwrite`: overwrite existing file?
@@ -809,21 +721,24 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                    data: str,
                    *,
                    overwrite: bool = False,
-                   encoding: str = 'utf-8',
-                   errors: str = 'strict',
-                   ) -> int:
-        '''
-        Return number of characters written.
-        '''
-        n = len(data)
-        z = data.encode(encoding=encoding, errors=errors)
+                   ) -> None:
+        z = data.encode(encoding='utf-8', errors='strict')
         self.write_bytes(z, overwrite=overwrite)
-        return n
 
 
-# Add methods 'read_json', 'write_json', 'read_pickle', 'write_pickle', etc.
+# Add methods
+# 'read_json', 'write_json', 'read_json_z', 'write_json_z', 'read_json_zstd', 'write_json_zstd',
+# 'read_pickle', 'write_pickle', 'read_pickle_z', 'write_pickle_z', 'read_pickle_zstd', 'write_pickle_zstd',
+# 'read_orjson', 'write_orjson', 'read_orjson_z', 'write_orjson_z', 'read_orjson_zstd', 'write_orjson_zstd',
+#
+# Applications can follow these examples to define and register their custom formats.
+
 Upath.register_read_write_text_format(JsonSerializer, 'json')
+Upath.register_read_write_byte_format(ZJsonSerializer, 'json_z')
+Upath.register_read_write_byte_format(ZstdJsonSerializer, 'json_zstd')
 Upath.register_read_write_byte_format(PickleSerializer, 'pickle')
-Upath.register_read_write_byte_format(CompressedPickleSerializer, 'pickle_z')
+Upath.register_read_write_byte_format(ZPickleSerializer, 'pickle_z')
+Upath.register_read_write_byte_format(ZstdPickleSerializer, 'pickle_zstd')
 Upath.register_read_write_byte_format(OrjsonSerializer, 'orjson')
-Upath.register_read_write_byte_format(CompressedOrjsonSerializer, 'orjson_z')
+Upath.register_read_write_byte_format(ZOrjsonSerializer, 'orjson_z')
+Upath.register_read_write_byte_format(ZstdOrjsonSerializer, 'orjson_zstd')
