@@ -1,12 +1,12 @@
+# User runs these tests with a `Upath` object of their subclass,
+# in a path that is safe for testing.
+
 import concurrent.futures
 import pathlib
 import time
 import pytest
 from upathlib import Upath, LocalUpath, LockAcquisitionTimeoutError
 
-# User runs these tests with a `Upath` object of their subclass,
-# in a path that is safe for testing.
-
 
 def test_basic(p: Upath):
     pp = p / '/abc/def/'
@@ -121,8 +121,9 @@ def test_read_write_rm_navigate(p: Upath):
     assert p3.remove_dir() == 0
     assert p.ls() == [p1]
     assert p1.remove_dir() == 0
-    assert p1.remove_file() == 1
-    assert p1.remove_file() == 0
+    p1.remove_file()
+    with pytest.raises(FileNotFoundError):
+        p1.remove_file()
     assert p.rmrf() == 0
 
 
@@ -169,6 +170,8 @@ def test_rename(p: Upath):
     (p / 'c/d/e.txt').write_text('e')
     (p / 'c/d.txt').write_text('d')
 
+    assert (p / 'a/a.txt').read_text() == 'a'
+
     p.joinpath('a/a.txt').rename_file('b/a.txt')
     assert not (p / 'a/a.txt').exists()
     assert (p / 'a/b/a.txt').read_text() == 'a'
@@ -190,7 +193,7 @@ def _access_in_mp(root: Upath, path: str, timeout):
         return t0 - time.perf_counter()
 
 
-def test_lock(p: Upath, timeout=None, wait=3):
+def test_lock(p: Upath, timeout=None, wait=8):
     p.rmrf()
     pp = p / 'testlock'
     with pp.lock(timeout=timeout):
@@ -198,7 +201,7 @@ def test_lock(p: Upath, timeout=None, wait=3):
             t = pool.submit(_access_in_mp, p / '/', pp._path, wait)
             z = t.result()
             print('mp returned after', z, 'seconds')
-            assert z <= -wait
+            assert z <= -(wait / 2)
 
 
 def test_all(p: Upath):
@@ -209,3 +212,5 @@ def test_all(p: Upath):
     test_read_write_rm_navigate(p)
     test_import_export(p)
     test_rename(p)
+
+    # test_lock(p)
